
from enum import Enum

class BetdaqError(Exception):
    """Base class for Betdaq Errors"""
    pass


class APIError(BetdaqError):
    """Exception raised if error is found"""

    def __init__(self, response, method=None, params=None, exception=None):
        if response:
            error_data = response.get('ResponseStatus', {})
            message = '%s \nParams: %s \nException: %s \nErrorCode: %s \nError Detail: %s' % (
                method, params, exception, error_data.get('Code'), response.get('ExtraInformation')
            )
        else:
            message = '%s \nParams: %s \nException: %s' % (
                method, params, exception
            )
        super(APIError, self).__init__(message)


class ResourceError(BetdaqError):
    def __init__(self):
        message = 'The API call was not processed successfully because of some critical resource constraint.' \
                  'This exception was as a result of a serious resource constraint within the exchange.'
        super(ResourceError, self).__init__(message)


class BetdaqSystemError(BetdaqError):
    def __init__(self):
        message = 'The API call was not processed successfully because of some serious technical error within the system.'
        super(BetdaqSystemError, self).__init__(message)


class EventClassifierDoesNotExist(BetdaqError):
    def __init__(self, event):
        message = 'An Event Classifier with the handle %s does not exist.' % event
        super(EventClassifierDoesNotExist, self).__init__(message)


class MarketDoesNotExist(BetdaqError):
    def __init__(self, market):
        message = 'A Market with the handle %s does not exist.' % market
        super(MarketDoesNotExist, self).__init__(message)


class SelectionDoesNotExist(BetdaqError):
    def __init__(self, selection):
        message = 'A selection with the handle %s does not exist.' % selection
        super(SelectionDoesNotExist, self).__init__(message)


class MarketNotActive(BetdaqError):
    def __init__(self, market):
        message = 'The action requested cannot be performed because the Market %s is not active.' % market
        super(MarketNotActive, self).__init__(message)


class MarketNeitherSuspendedNorActive(BetdaqError):
    def __init__(self, market):
        message = 'The action requested cannot be performed because the Market %s is not active or suspended.' % market
        super(MarketNeitherSuspendedNorActive, self).__init__(message)


class SelectionNotActive(BetdaqError):
    def __init__(self, selection):
        message = 'The action requested cannot be performed because the selection %s is not active or suspended.' % selection
        super(SelectionNotActive, self).__init__(message)


class InsufficientVirtualPunterFunds(BetdaqError):
    def __init__(self):
        message = 'The Order specified could not be placed because the increase in the amount of the Virtual ' \
                  'Punter’s funds that would need to be frozen as a result of that Order is greater than the ' \
                  "value specified for ‘maxVirtualReservationIncrease’"
        super(InsufficientVirtualPunterFunds, self).__init__(message)


class OrderDoesNotExist(BetdaqError):
    def __init__(self, order_id):
        message = 'An Order with the handle %s does not exist.' % order_id
        super(OrderDoesNotExist, self).__init__(message)


class NoUnmatchedAmount(BetdaqError):
    def __init__(self, order_id):
        message = 'The Order {} could not be placed, cancelled or changed because the amount ' \
                  'requested is negative or the entire stake of the Order has already been matched.'.format(order_id)
        super(NoUnmatchedAmount, self).__init__(message)


class ResetHasOccurred(BetdaqError):
    def __init__(self, order_id):
        message = 'The order was not placed because the expectedSelectionResetCount {} on the Order does not match ' \
                  'the current selectionResetCount for the Selection.'.format(order_id)
        super(ResetHasOccurred, self).__init__(message)


class OrderAlreadySuspended(BetdaqError):
    def __init__(self, order_id):
        message = 'The order %s is already suspended' % order_id
        super(OrderAlreadySuspended, self).__init__(message)


class TradingCurrentlySuspended(BetdaqError):
    def __init__(self, order_id):
        message = 'The order %s could not be processed because all trading is currently suspended on the Exchange.' % order_id
        super(TradingCurrentlySuspended, self).__init__(message)


class InvalidOdds(BetdaqError):
    def __init__(self, odds, stake):
        message = 'The odds %s or stake %s are not valid.' % (odds, stake)
        super(InvalidOdds, self).__init__(message)


class WithdrawalSequenceNumberIsInvalid(BetdaqError):
    def __init__(self, sequenceNumber):
        message = 'The Order was not placed because the withdrawal sequence number %s is greater than the current' \
                  ' withdrawal sequence number for the market.' % sequenceNumber
        super(WithdrawalSequenceNumberIsInvalid, self).__init__(message)


class MaximumInputRecordsExceeded(BetdaqError):
    def __init__(self):
        message = 'Exceeded the limit on the number of parameters that can be specified.'
        super(MaximumInputRecordsExceeded, self).__init__(message)


class PunterSuspended(BetdaqError):
    def __init__(self):
        message = 'User is currently suspended.'
        super(PunterSuspended, self).__init__(message)


class PunterProhibitedFromPlacingOrders(BetdaqError):
    def __init__(self):
        message = 'Punter is prohibited from placing orders on exchange.'
        super(PunterProhibitedFromPlacingOrders, self).__init__(message)


class InsufficientPunterFunds(BetdaqError):
    def __init__(self):
        message = 'Punter does not have sufficient unfrozen funds.'
        super(InsufficientPunterFunds, self).__init__(message)


class OrderAPIInProgress(BetdaqError):
    def __init__(self):
        message = 'Another order API is currently in progress for this Punter. A Punter can not issue more ' \
                  'than one order API (PlaceSimgleOrder, PlaceGroupOrder or ChangeOrder) at the same time.'
        super(OrderAPIInProgress, self).__init__(message)


class PunterOrderMismatch(BetdaqError):
    def __init__(self):
        message = 'One or more of the Orders specified were not issued by the Punter specified or there is a ' \
                  'mis-match between the Punter implied and the object specified.'
        super(PunterOrderMismatch, self).__init__(message)


class MarketNotEnabledForMultiples(BetdaqError):
    def __init__(self):
        message = 'The market of one or more selections does not support multiple bets.	'
        super(MarketNotEnabledForMultiples, self).__init__(message)


class MultipleLayerParameterAlreadyExists(BetdaqError):
    def __init__(self):
        message = 'A MultiplePriceMultipler for the MultipleLayer with the numberOfSelections specified already exists.'
        super(MultipleLayerParameterAlreadyExists, self).__init__(message)


class LevelsRequestedExceedsMaximum(BetdaqError):
    def __init__(self):
        message = 'The number of levels of combination requested exceeds the system defined maximum'
        super(LevelsRequestedExceedsMaximum, self).__init__(message)


class NoMultipleOfferAvailable(BetdaqError):
    def __init__(self):
        message = 'The requested multiple bet was not matched because there was no multiple offers available.'
        super(NoMultipleOfferAvailable, self).__init__(message)


class InRunningDelayInEffect(BetdaqError):
    def __init__(self):
        message = 'An attempt was made to change an order that is currently subject to an in-running delay.'
        super(InRunningDelayInEffect, self).__init__(message)


class MultipleSelectionsUnderSameEvent(BetdaqError):
    def __init__(self):
        message = 'An attempt was made to place a multiple bet that contained two or more ' \
                  'Selections belonging to markets under the same event classifier'
        super(MultipleSelectionsUnderSameEvent, self).__init__(message)


class MultipleSelectionsWithSameName(BetdaqError):
    def __init__(self):
        message = 'An attempt was made to place a multiple bet that contained two or more selections ' \
                  'with exactly the same name.'
        super(MultipleSelectionsWithSameName, self).__init__(message)


class DuplicateOrderSpecified(BetdaqError):
    def __init__(self):
        message = 'The same order was specified more than once in the same API call.'
        super(DuplicateOrderSpecified, self).__init__(message)


class OrderNotSuspended(BetdaqError):
    def __init__(self):
        message = 'The order is not currently suspended.'
        super(OrderNotSuspended, self).__init__(message)


class PunterIsSuspendedFromTrading(BetdaqError):
    def __init__(self):
        message = 'Punter concerned is currently suspended from trading.'
        super(PunterIsSuspendedFromTrading, self).__init__(message)


class PunterHasActiveOrders(BetdaqError):
    def __init__(self):
        message = 'An attempt was made to unsuspend Punter from trading. A Punter can only be ' \
                  'unsuspended if the Punter currently has no active orders. ' \
                  'However the Punter concerned currently does have active orders.'
        super(PunterHasActiveOrders, self).__init__(message)


class PunterNotSuspendedFromTrading(BetdaqError):
    def __init__(self):
        message = 'The requested operation could not be completed because the Punter concerned is not currently suspended from trading'
        super(PunterNotSuspendedFromTrading, self).__init__(message)


class ExpiryTimeInThePast(BetdaqError):
    def __init__(self):
        message = 'The requested operation could not be performed because the expiry time specified is in the past'
        super(ExpiryTimeInThePast, self).__init__(message)


class NoChangeSpecified(BetdaqError):
    def __init__(self):
        message = 'Change order could not be performed because the price specified is the current price of the order concerned'
        super(NoChangeSpecified, self).__init__(message)


class SoapHeaderNotSupplied(BetdaqError):
    def __init__(self):
        message = 'The SOAP header was not specified. All External API calls must include a SOAP header.'
        super(SoapHeaderNotSupplied, self).__init__(message)


class IncorrectVersionNumber(BetdaqError):
    def __init__(self):
        message = 'An incorrect version number specified in API header.'
        super(IncorrectVersionNumber, self).__init__(message)


class NoUsernameSpecified(BetdaqError):
    def __init__(self):
        message = 'You must specify your username in the API header.'
        super(NoUsernameSpecified, self).__init__(message)


class InvalidParameters(BetdaqError):
    def __init__(self):
        message = 'Invalid parameters were passed to the web method.'
        super(InvalidParameters, self).__init__(message)


class NoPasswordSpecified(BetdaqError):
    def __init__(self):
        message = 'You must specify your password in the API header.'
        super(NoPasswordSpecified, self).__init__(message)


class MultipleCombinationExclusionAlreadyExists(BetdaqError):
    def __init__(self):
        message = 'There is currently an active MultipleCombinationExclusion with the set of Selections specified.'
        super(MultipleCombinationExclusionAlreadyExists, self).__init__(message)


class MultipleCombinationExlcusionDoesNotExist(BetdaqError):
    def __init__(self):
        message = 'There is not currently an active MultipleCombinationExclusion with the set of Selections specified ' \
                  'for the multiple layer specified.'
        super(MultipleCombinationExlcusionDoesNotExist, self).__init__(message)


class InvalidPassword(BetdaqError):
    def __init__(self):
        message = 'The password specified is not a valid password. Specifically it does not conform to the rules ' \
                  'defined for such passwords.'
        super(InvalidPassword, self).__init__(message)


class PunterIsBlacklisted(BetdaqError):
    def __init__(self):
        message = 'The operation requested was not executed because the Punter concerned is currently ' \
                  'black-listed from performing that action.'
        super(PunterIsBlacklisted, self).__init__(message)


class PunterNotRegisteredAsMultipleLayer(BetdaqError):
    def __init__(self):
        message = 'The requested action could not be performed because the Punter concerned has not been ' \
                  'registered as a multiple layer.'
        super(PunterNotRegisteredAsMultipleLayer, self).__init__(message)


class PunterAlreadyRegisteredForHeartbeat(BetdaqError):
    def __init__(self):
        message = 'The operation could not complete because the Punter concerned is already registered for a Heartbeat.'
        super(PunterAlreadyRegisteredForHeartbeat, self).__init__(message)


class PunterNotRegisteredForHeartbeat(BetdaqError):
    def __init__(self):
        message = 'The operation could not complete because the Punter concerened is not registered for a Heartbeat.'
        super(PunterNotRegisteredForHeartbeat, self).__init__(message)


class ThresholdSpecifiedTooSmall(BetdaqError):
    def __init__(self):
        message = 'The threshold value specified is less than the system defined minimum threshold value.'
        super(ThresholdSpecifiedTooSmall, self).__init__(message)


class UnmatchedOrderCouldResult(BetdaqError):
    def __init__(self):
        message = 'The attempt to place an order was not successful because it could result in an unmatched order ' \
                  'remaining on the system. The specific API call used can only be used for orders that can not result' \
                  ' in a an unmatched order. For example, you can not specify a KillType of FillOrKillDontCancel.'
        super(UnmatchedOrderCouldResult, self).__init__(message)


class PunterNotAuthorisedForAPI(BetdaqError):
    def __init__(self):
        message = 'The punter concerned is not authorised to use the external API.'
        super(PunterNotAuthorisedForAPI, self).__init__(message)


class MarketIsForRealMoney(BetdaqError):
    def __init__(self):
        message = 'The requested operation cannot be performed because the market concerned is not a play market ' \
                  'but the currency specified is a play currency.'
        super(MarketIsForRealMoney, self).__init__(message)


class MarketIsForPlayMoney(BetdaqError):
    def __init__(self):
        message = 'The requested operation cannot be performed because the market concerned is a plat market ' \
                  'but the currency specified is a real currency.'
        super(MarketIsForPlayMoney, self).__init__(message)


class CannotChangeToSPIfUnmatched(BetdaqError):
    def __init__(self):
        message = 'The requested operation could not be performed because the orderFillType of the order ' \
                  'concerned is not Normal.'
        super(CannotChangeToSPIfUnmatched, self).__init__(message)


class InvalidUsername(BetdaqError):
    def __init__(self):
        message = 'Username provided is not valid.'
        super(InvalidUsername, self).__init__(message)




import bisect

from betdaq.utils import make_tz_naive, price_side_map
from betdaq.enums import MarketType, MarketStatus, SelectionStatus, Polarity


def parse_deep_markets(sports):
    markets = []
    for sport in sports:
        events = sport.get('EventClassifiers')
        sub_events = []
        while events:
            for event in events:
                markets += parse_market(
                    event.get('Markets', []), {'event_name': event.get('Name'),
                                               'tournament_id': event.get('tournament_id'),
                                               'tournament_name': event.get('tournament_name'),
                                               'competition_id': event.get('competition_id'),
                                               'competition_name': event.get('competition_name'),
                                               'sport_id': sport.get('Id'),
                                               'sport_name': sport.get('Name'),
                                               }
                )
                if event.get('EventClassifiers', []):
                    sub_events += [{**ev, **{'competition_name': event.get('Name')
                                             if event.get('tournament_name') and event.get('competition_name') is None
                                             else event.get('competition_name') if event.get('competition_name')
                                             else None,
                                             'competition_id': event.get('Id')
                                             if event.get('tournament_name') and event.get('competition_name') is None
                                             else event.get('competition_id') if event.get('competition_name')
                                             else None,
                                             'tournament_name': event.get('tournament_name', event.get('Name')),
                                             'tournament_id': event.get('tournament_id', event.get('Id')),
                                             }
                                    } for ev in event.get('EventClassifiers', [])]
            events = sub_events.copy()
            sub_events = []
    return markets


def parse_runners(data):
    return {'runner_id': data.get('Id'),
            'runner_name': data.get('Name'),
            'runner_status': SelectionStatus(int(data.get('Status'))).name if data.get('Status') else None,
            'reset_count': data.get('ResetCount'),
            'deduction_factor': float(data.get('DeductionFactor')),
            'runner_display_order': data.get('DisplayOrder')}


def parse_market(mkt_data, other_info):
    return [{**{'runners': [parse_runners(runner) for runner in mkt.get('Selections', [])],
                'market_id':  mkt.get('Id'),
                'market_name':  mkt.get('Name'),
                'market_type':  MarketType(int(mkt.get('Type'))).name if mkt.get('Type') else None,
                'is_play_market': mkt.get('IsPlayMarket'),
                'market_status': MarketStatus(int(mkt.get('Status'))).name if mkt.get('Status') else None,
                'number_of_winners': mkt.get('NumberOfWinningSelections'),
                'market_start_time': make_tz_naive(mkt.get('StartTime')),
                'withdrawal_sequence_number': mkt.get('WithdrawalSequenceNumber'),
                'market_display_order': mkt.get('DisplayOrder'),
                'enabled_for_multiples': mkt.get('IsEnabledForMultiples'),
                'in_play_available': mkt.get('IsInRunningAllowed'),
                'race_grade': mkt.get('RaceGrade'),
                'managed_in_running': mkt.get('IsManagedWhenInRunning'),
                'in_play': mkt.get('IsCurrentlyInRunning'),
                'in_play_delay': mkt.get('InRunningDelaySeconds'),
                'event_id': mkt.get('EventClassifierId'),
                'place_payout': mkt.get('PlacePayout')},
             **other_info} for mkt in mkt_data]


def parse_sub_event(data, prefix='comp', parent='sport'):
    return {
        '%s_id' % prefix: data.get('Id'),
        '%s_name' % prefix: data.get('Name'),
        '%s_display_order' % prefix: data.get('DisplayOrder'),
        '%s_id' % parent: data.get('ParentId'),
        '%s_multi_allowed' % prefix: data.get('IsEnabledForMultiples'),
    }


def parse_market_prices(mkt):
    return {'market_id': mkt.get('Id'),
            'market_name': mkt.get('Name'),
            'market_type': MarketType(int(mkt.get('Type'))).name if mkt.get('Type') else None,
            'market_start_time': make_tz_naive(mkt.get('StartTime')),
            'runners': [parse_runner_prices(runner) for runner in mkt.get('Selections', [])],
            'is_play_market': mkt.get('IsPlayMarket'),
            'status': MarketStatus(int(mkt.get('Status'))) if mkt.get('Status') else None,
            'number_of_winners': mkt.get('NumberOfWinningSelections'),
            'withdrawal_sequence_number': mkt.get('WithdrawalSequenceNumber'),
            'market_display_order': mkt.get('DisplayOrder'),
            'enabled_for_multiples': mkt.get('IsEnabledForMultiples'),
            'in_play_available': mkt.get('IsInRunningAllowed'),
            'race_grade': mkt.get('RaceGrade'),
            'managed_in_running': mkt.get('IsManagedWhenInRunning'),
            'in_play': mkt.get('IsCurrentlyInRunning'),
            'in_running_delay': mkt.get('InRunningDelaySeconds'),
            'event_id': mkt.get('EventClassifierId'),
            'market_total_matched': mkt.get('TotalMatchedAmount'),
            'place_payout': mkt.get('PlacePayout'),
            'market_back_matched': mkt.get('MatchedMarketForStake'),
            'market_lay_matched': mkt.get('MatchedMarketAgainstStake'),
            'home_team_score': mkt.get('HomeTeamScore'),
            'away_team_score': mkt.get('AwayTeamScore'),
            'score_type': mkt.get('ScoreType'),
            }


def parse_runner_prices(runner):
    return {
        'runner_book': parse_runner_book(runner.get('_value_1', [])),
        'runner_id': runner.get('Id'),
        'runner_name': runner.get('Name'),
        'runner_status': SelectionStatus(int(runner.get('Status'))).name if runner.get('Status') else None,
        'runner_reset_count': runner.get('ResetCount'),
        'deduction_factor': runner.get('DeductionFactor'),
        'runner_back_matched_size': runner.get('MatchedSelectionForStake'),
        'runner_lay_matched_size': runner.get('MatchedSelectionAgainstStake'),
        'runner_last_matched_time': make_tz_naive(runner.get('LastMatchedOccurredAt')),
        'runner_last_matched_price': runner.get('LastMatchedPrice'),
        'runner_last_matched_back_size': runner.get('LastMatchedForSideAmount'),
        'runner_last_matched_lay_size': runner.get('LastMatchedAgainstSideAmount'),
        'runner_open_interest': runner.get('SelectionOpenInterest'),
        'runner_market_winnings': runner.get('MarketWinnings'),
        'runner_positive_winnings': runner.get('MarketPositiveWinnings'),
        'runner_back_matched_same_price': runner.get('MatchedForSideAmountAtSamePrice'),
        'runner_lay_matched_same_price': runner.get('MatchedAgainstSideAmountAtSamePrice'),
        'runner_last_traded_same_price': make_tz_naive(runner.get('FirstMatchAtSamePriceOccurredAt')),
        'runner_total_matched_orders': runner.get('NumberOrders'),
        'runner_total_matched_punters': runner.get('NumberPunters'),
        }


def parse_runner_book(book):
    back_levels = []
    lay_levels = []
    order_book = {'batb': [], 'batl': []}
    for level in book:
        for side, order in level.items():
            if order:
                side = price_side_map.get(side)
                if side == 'back':
                    bisect.insort(back_levels, float(order.get('Price')))
                    order_book['batb'].append([float(order.get('Price')), float(order.get('Stake'))])
                elif side == 'lay':
                    bisect.insort_right(lay_levels, float(order.get('Price')))
                    order_book['batl'].append([float(order.get('Price')), float(order.get('Stake'))])
    back_levels.reverse()
    order_book['batb'] = [[back_levels.index(x[0]), x[0], x[1]] for x in order_book['batb']]
    order_book['batl'] = [[lay_levels.index(x[0]), x[0], x[1]] for x in order_book['batl']]
    return order_book


def parse_selection_changes(chg):
    return {
        'runner_id': chg.get('Id'),
        'runner_name': chg.get('Name'),
        'runner_display_order': chg.get('DisplayOrder'),
        'runner_hidden': chg.get('IsHidden'),
        'runner_status': SelectionStatus(int(chg.get('Status'))) if chg.get('Status') else None,
        'reset_count': chg.get('ResetCount'),
        'market_id': chg.get('MarketId'),
        'withdrawal_factor': chg.get('WithdrawalFactor'),
        'sequence_number': chg.get('SelectionSequenceNumber'),
        'cancel_orders_time': make_tz_naive(chg.get('CancelOrdersTime')),
        'settlement_info': [{'settled_time': make_tz_naive(stl.get('SettlementInformation', {}).get('SettledTime')),
                             'void_percentage': stl.get('SettlementInformation', {}).get('VoidPercentage'),
                             'result': stl.get('SettlementInformation', {}).get('SettlementResultString'),
                             'left_side_percentage': stl.get('SettlementInformation', {}).get('LeftSideFactor'),
                             'right_side_percentage': stl.get('SettlementInformation', {}).get('RightSideFactor')
                             } for stl in chg.get('_value_1', [])]
    }


def parse_market_withdrawal(data):
    return {
        'runner_id': data.get('SelectionId'),
        'withdrawal_time': make_tz_naive(data.get('WithdrawalTime')),
        'sequence_number': data.get('SequenceNumber'),
        'reduction_factor': data.get('ReductionFactor'),
        'compound_reduction_factor': data.get('CompoundReductionFactor'),
    }


def parse_ladder(data):
    return [
        {'price': float(ol.get('price')),
         'value': ol.get('representation')} for ol in data.get('Ladder', {})]


def parse_sports(sport):
    return {'display_order': sport.get('DisplayOrder'),
            'sport_id': sport.get('Id'),
            'sport_name': sport.get('Name')}


def parse_trade_item(trade):
    trd = trade.get('TradeItems', {})
    return {
        'traded_time': make_tz_naive(trd.get('occurredAt')),
        'price': trd.get('price'),
        'size': trd.get('backersStake'),
        'side': Polarity(int(trd.get('tradeType'))).name if trd.get('tradeType') else None,
    }


def parse_selection_trades(trades):
    return {
        'runner_id': trades.get('selectionId'),
        'max_trade_id': trades.get('maxTradeId'),
        'max_trade_id_returned': trades.get('maxTradeIdReturned'),
        'trades': [parse_trade_item(t) for t in trades.get('_value_1', [])],
    }




import datetime

from betdaq.enums import Boolean
from betdaq.utils import clean_locals, listy_mc_list
from betdaq.endpoints.baseendpoint import BaseEndpoint
from betdaq.errorparsers.betting import err_cancel_market, err_suspend_orders
from betdaq.resources.bettingresources import (
    parse_orders, parse_single_order, parse_orders_receipt, parse_order_update,
    parse_cancelled_order, parse_suspended_order,
)


class Betting(BaseEndpoint):

    def get_orders(self, SequenceNumber=-1, wantSettledOrdersOnUnsettledMarkets=Boolean.T.value):
        """
        Get the initial list of orders that's need to be taken into consideration when establishing positions. 
        Information about the following orders will be returned:
            •	active orders
            •	fully matched orders
            •	cancelled orders that have a matched portion 
            •	suspended orders
            •	some settled or voided orders under some conditions

        :param SequenceNumber: lower bound cutoff for sequence updates to include, 
                               -1 will set to earliest possible sequence.
        :type SequenceNumber: int
        :param wantSettledOrdersOnUnsettledMarkets: Flag indicating whether or not information about settled orders 
                                                    on unsettled markets should be returned.
        :type wantSettledOrdersOnUnsettledMarkets: bool
        :return: orders that have changed.
        """
        params = clean_locals(locals())
        date_time_sent = datetime.datetime.utcnow()
        response = self.request('ListBootstrapOrders', params, secure=True)
        data = self.process_response(response, date_time_sent, 'Orders')
        return [parse_orders(order) for order in data.get('data', {}).get('Order', [])] if data.get('data') else []

    def get_orders_diff(self, SequenceNumber):
        """
        Get a list of orders for the logged in user that have changed since a given sequence number.
        Utilised to maintain position information after initial position is established with list_orders.

        :param SequenceNumber: lower bound cutoff for sequence updates to include.
        :type SequenceNumber: int
        :return: orders that have changed.
        """
        params = clean_locals(locals())
        date_time_sent = datetime.datetime.utcnow()
        response = self.request('ListOrdersChangedSince', params, secure=True)
        data = self.process_response(response, date_time_sent, 'Orders')
        return [parse_orders(order) for order in data.get('data', {}).get('Order', [])] if data.get('data') else []

    def get_single_order(self, OrderId):
        """
        Get full detail and history about an individual order.

        :param order_id: id of the order we wish to get history for.
        :type order_id: int
        :return: single orders history and current status.
        """
        params = clean_locals(locals())
        date_time_sent = datetime.datetime.utcnow()
        response = self.request('GetOrderDetails', params, secure=True)
        data = self.process_response(response, date_time_sent, None)
        return listy_mc_list(parse_single_order(data.get('data', {})) if data.get('data') else {})

    def place_orders(self, order_list, WantAllOrNothingBehaviour=Boolean.T.value, receipt=True):
        """
        Places one or more orders at exchange. 
        Receipt determines whether to wait for complete matching cycle or just return the order ID.
        
        :param order_list: List of orders to be sent to exchange
        :type order_list: list of betdaq_py.filters.create_order
        :param WantAllOrNothingBehaviour: defines whether to kill all orders on any error or place orders independently.
        :type WantAllOrNothingBehaviour: betdaq_py.enums.Boolean
        :param receipt: whether to wait for matching cycle and return full info of order or not.
        :type receipt: bool
        :return: the order ID or full order information depending on receipt.
        """
        date_time_sent = datetime.datetime.utcnow()
        if receipt:
            method = 'PlaceOrdersWithReceipt'
            params = self.client.secure_types['%sRequest' % method](Orders={'Order': order_list})
        else:
            method = 'PlaceOrdersNoReceipt'
            params = self.client.secure_types['%sRequest' % method](
                WantAllOrNothingBehaviour=WantAllOrNothingBehaviour, Orders={'Order': order_list}
            )
        response = self.request(method, params, secure=True)
        data = self.process_response(response, date_time_sent, 'Orders')
        return [
            parse_orders_receipt(order) for order in data.get('data', {}).get('Order', [])
        ] if data.get('data') else []

    def update_orders(self, order_list):
        """
        Update orders on exchange
        
        :param order_list: list of order updates to be sent to exchange.
        :type order_list: list of betdaq_py.filters.update_order
        :return: BetID and status of update.
        """
        params = self.client.secure_types['UpdateOrdersNoReceiptRequest'](Orders={'Order': order_list})
        date_time_sent = datetime.datetime.utcnow()
        response = self.request('UpdateOrdersNoReceipt', params, secure=True)
        data = self.process_response(response, date_time_sent, 'Orders')
        return [
            parse_order_update(update) for update in data.get('data', {}).get('Order', [])
        ] if data.get('data') else []

    def cancel_orders(self, order_ids):
        """
        Cancel one or more orders on exchange
        
        :param order_ids: list of order ids to be cancelled.
        :type order_ids: list of ints
        :return: information on the cancellation status of each order.
        """
        params = self.client.secure_types['CancelOrdersRequest'](
            _value_1=[{'OrderHandle': order_id} for order_id in order_ids]
        )
        date_time_sent = datetime.datetime.utcnow()
        response = self.request('CancelOrders', params, secure=True)
        data = self.process_response(response, date_time_sent, 'Orders')
        return [
            parse_cancelled_order(cancel) for cancel in data.get('data', {}).get('Order', [])
        ] if data.get('data') else []

    def cancel_orders_by_market(self, market_ids):
        """
        Cancel all orders on one or more markets on exchange
        
        :param market_ids: list of market ids to be cancelled.
        :type market_ids: list of ints
        :return: information on the cancellation status of each order.
        """
        params = self.client.secure_types['CancelAllOrdersOnMarketRequest'](
            _value_1=[{'MarketIds': market} for market in market_ids]
        )
        date_time_sent = datetime.datetime.utcnow()
        response = self.request('CancelAllOrdersOnMarket', params, secure=True)
        data = self.process_response(response, date_time_sent, 'Order', error_handler=err_cancel_market)
        return [parse_cancelled_order(cancel) for cancel in data.get('data', [])] if data.get('data') else []

    def cancel_all_orders(self):
        """
        Cancels all unmatched orders across all markets.

        :return: information on the cancellation status of each order.
        """
        date_time_sent = datetime.datetime.utcnow()
        response = self.request('CancelAllOrders', {}, secure=True)
        data = self.process_response(response, date_time_sent, 'Orders')
        return [
            parse_cancelled_order(cancel) for cancel in data.get('data', {}).get('Order', [])
        ] if data.get('data') else []

    def suspend_orders(self, order_ids):
        """
        Suspend one or more orders on exchange

        :param order_ids: list of order ids to be suspended.
        :type order_ids: list of ints
        :return: information on the suspension status of each order.
        """
        params = self.client.secure_types['SuspendOrdersRequest'](
            _value_1=[{'OrderIds': order_id} for order_id in order_ids]
        )
        date_time_sent = datetime.datetime.utcnow()
        response = self.request('SuspendOrders', params, secure=True)
        data = self.process_response(response, date_time_sent, 'Orders', error_handler=err_suspend_orders)
        return [parse_suspended_order(suspend) for suspend in data.get('data', [])] if data.get('data') else []

    def suspend_orders_by_market(self, MarketId):
        """
        Suspend all orders on a given market.

        :param MarketIds: market id to be suspend orders on.
        :type MarketIds: ints
        :return: information on the suspension status of each order.
        """
        params = clean_locals(locals())
        date_time_sent = datetime.datetime.utcnow()
        response = self.request('SuspendAllOrdersOnMarket', params, secure=True)
        data = self.process_response(response, date_time_sent, 'Orders', error_handler=err_suspend_orders)
        return [parse_suspended_order(suspend) for suspend in data.get('data', [])] if data.get('data') else []

    def suspend_all_orders(self):
        """
        Suspend all orders across all markets.

        :return: information on the suspension status of each order.
        """
        date_time_sent = datetime.datetime.utcnow()
        response = self.request('SuspendAllOrders', {}, secure=True)
        data = self.process_response(response, date_time_sent, 'Orders', error_handler=err_suspend_orders)
        return [parse_suspended_order(suspend) for suspend in data.get('data', [])] if data.get('data') else []

    def unsuspend_orders(self, order_ids):
        """
        Unsuspends one or more suspended orders
        
        :param order_ids: list of order ids to unsuspend.
        :type order_ids: list
        :return: 
        """
        params = self.client.secure_types['UnsuspendOrdersRequest'](
            _value_1=[{'OrderIds': order_id} for order_id in order_ids]
        )
        date_time_sent = datetime.datetime.utcnow()
        response = self.request('UnsuspendOrders', params, secure=True)
        data = self.process_response(response, date_time_sent, None)
        return []




from betdaq.utils import make_tz_naive
from betdaq.enums import OrderActionType, OrderStatus, OrderKillType, Polarity, MarketStatus


def parse_suspended_order(suspend):
    return {
        'order_id': suspend.get('OrderId'),
        'size_suspended': suspend.get('SuspendedForSideStake'),
        'customer_reference': suspend.get('PunterReferenceNumber'),
    }


def parse_cancelled_order(cancel):
    return {
        'order_id': cancel.get('OrderHandle'),
        'size_cancelled': cancel.get('cancelledForSideStake'),
        'customer_reference': cancel.get('PunterReferenceNumber'),
    }


def parse_order_update(update):
    return {
        'order_id': update.get('BetId'),
        'return_code': update.get('ReturnCode'),
    }


def parse_orders_receipt(order):
    return {
        'order_id': order.get('OrderHandle'),
        'side': Polarity(int(order.get('Polarity'))).name if order.get('Polarity') else None,
        'size_remaining': order.get('UnmatchedStake'),
        'matched_price': order.get('MatchedPrice'),
        'matched_size': order.get('MatchedStake'),
        'matched_lay_size': order.get('MatchedAgainstStake'),
        'sent_time': make_tz_naive(order.get('IssuedAt')),
        'status': OrderStatus(int(order.get('Status'))) if order.get('Status') else None,
        'runner_sequence_number': order.get('SequenceNumber'),
        'runner_id': order.get('SelectionId'),
        'customer_reference': order.get('PunterReferenceNumber'),
        'return_code': order.get('ReturnCode'),
    }


def parse_orders(order):
    return {
        'order_id': order.get('Id'),
        'commission_information': {
                'gross_settlement_amount': order.get('OrderCommissionInformation', {}).get('GrossSettlementAmount'),
                'commission': order.get('OrderCommissionInformation', {}).get('OrderCommission'),
        } if order.get('OrderCommissionInformation') else {},
        'runner_id': order.get('SelectionId'),
        'market_id': order.get('MarketId'),
        'sequence_number': order.get('SequenceNumber'),
        'status': OrderStatus(int(order.get('Status'))).name if order.get('Status') else None,
        'side': Polarity(int(order.get('Polarity'))).name if order.get('Polarity') else None,
        'sent_time': make_tz_naive(order.get('IssuedAt')),
        'price': order.get('RequestedPrice'),
        'remaining_size': order.get('UnmatchedStake'),
        'average_price': order.get('AveragePrice'),
        'matched_price': order.get('MatchedPrice'),
        'matched_size': order.get('MatchedStake'),
        'matched_lay_size': order.get('MatchedAgainstStake'),
        'back_take_size': order.get('TotalForSideTakeStake '),
        'back_make_size': order.get('TotalForSideMakeStake '),
        'customer_reference': order.get('PunterReferenceNumber'),
        'withdrawal_sequence_number': order.get('ExpectedWithdrawalSequenceNumber'),
        'runner_reset_count': order.get('ExpectedSelectionResetCount'),
        'in_play': order.get('IsCurrentlyInRunning'),
        'order_fill_type': OrderKillType(int(order.get('OrderFillType'))).name if order.get('OrderFillType') else None,
        'fill_or_kill_threshold': order.get('FillOrKillThreshold'),
        'cancel_on_in_running': order.get('CancelOnInRunning'),
        'cancel_if_selection_reset': order.get('CancelIfSelectionReset'),
        'commission_base_rate': order.get('PunterCommissionBasis'),
        'commission_take_rate': order.get('TakeCommissionRate'),
        'commission_make_rate': order.get('MakeCommissionRate'),
        }


def parse_single_order(order):
    return {
        'order_id': order.get('Id'),
        'settlement_information': {
                'gross_settlement_amount': order.get('OrderCommissionInformation', {}).get('GrossSettlementAmount')
                if order.get('OrderCommissionInformation') else None,
                'order_commission': order.get('OrderCommissionInformation', {}).get('OrderCommission')
                if order.get('OrderCommissionInformation') else None,
                'market_commission': order.get('OrderCommissionInformation', {}).get('MarketCommission')
                if order.get('OrderCommissionInformation') else None,
                'settlement_time': make_tz_naive(order.get('OrderCommissionInformation', {}).get('MarketSettledDate'))
                if order.get('OrderCommissionInformation') else None,
        },
        'audit_log': [{'timestamp': make_tz_naive(log.get('Time')),
                       'matched_size': log.get('MatchedOrderInformation', {}).get('MatchedStake')
                       if log.get('MatchedOrderInformation') else None,
                       'matched_lay_size': log.get('MatchedOrderInformation', {}).get('MatchedAgainstStake')
                       if log.get('MatchedOrderInformation') else None,
                       'matched_price': log.get('MatchedOrderInformation', {}).get('PriceMatched')
                       if log.get('MatchedOrderInformation') else None,
                       'matched_order_id': log.get('MatchedOrderInformation', {}).get('MatchedOrderID')
                       if log.get('MatchedOrderInformation')else None,
                       'maker': log.get('MatchedOrderInformation', {}).get('WasMake')
                       if log.get('MatchedOrderInformation') else None,
                       'order_commission': log.get('CommissionInformation', {}).get('OrderCommission'),
                       'gross_settlement_amount': log.get('CommissionInformation', {}).get('GrossSettlementAmount'),
                       'order_action': OrderActionType(int(log.get('OrderActionType'))).name,
                       'requested_size': log.get('RequestedStake'),
                       'total_size': log.get('TotalStake'),
                       'total_lay_size': log.get('TotalAgainstStake'),
                       'requested_price': log.get('RequestedPrice'),
                       'average_price': log.get('AveragePrice'),
                       } for log in order.get('AuditLog', {}).get('AuditLog', [])],
        'runner_id': order.get('SelectionId'),
        'market_id': order.get('MarketId'),
        'status': OrderStatus(int(order.get('OrderStatus'))).name if order.get('OrderStatus') else None,
        'in_play': order.get('IsCurrentlyInRunning'),
        'market_status': MarketStatus(int(order.get('MarketStatus'))).name if order.get('MarketStatus') else None,
        'sent_time': make_tz_naive(order.get('IssuedAt')),
        'last_update_time': make_tz_naive(order.get('LastChangedAt')),
        'expiry_time': make_tz_naive(order.get('ExpiresAt')),
        'valid_from': make_tz_naive(order.get('ValidFrom')),
        'order_fill_type': OrderKillType(int(order.get('OrderFillType'))).name if order.get('OrderFillType') else None,
        'fill_or_kill_threshold': order.get('FillOrKillThreshold'),
        'requested_size': order.get('RequestedStake'),
        'requested_price': order.get('RequestedPrice'),
        'expected_selection_reset_count': order.get('ExpectedSelectionResetCount'),
        'total_size': order.get('TotalStake'),
        'remaining_size': order.get('UnmatchedStake'),
        'matched_size': order.get('MatchedStake'),
        'matched_lay_stake': order.get('MatchedAgainstStake'),
        'matched_price': order.get('MatchedPrice'),
        'average_price': order.get('AveragePrice'),
        'matching_time': make_tz_naive(order.get('MatchingTimeStamp')),
        'side': Polarity(int(order.get('Polarity'))) if order.get('Polarity') else None,
        'withdrawal_reprice_option': order.get('WithdrawalRepriceOption'),
        'cancel_on_in_running': order.get('CancelOnInRunning'),
        'cancel_if_selection_reset': order.get('CancelIfSelectionReset'),
        'sequence_number': order.get('SequenceNumber'),
        'customer_reference': order.get('PunterReferenceNumber'),
        'market_type': order.get('MarketType'),
        'expected_withdrawal_sequence_number': order.get('ExpectedWithdrawalSequenceNumber'),
        }




from betdaq.utils import make_tz_naive


def parse_account_postings(posts):
    return {
            'postings_complete': posts.get('HaveAllPostingsBeenReturned'),
            'currency': posts.get('Currency'),
            'available_funds': posts.get('AvailableFunds'),
            'balance': posts.get('Balance'),
            'credit': posts.get('Credit'),
            'exposure': posts.get('Exposure'),
            'transactions': [{
                'transaction_time': make_tz_naive(order.get('PostedAt')),
                'description': order.get('Description'),
                'amount': order.get('Amount'),
                'resulting_balance': order.get('ResultingBalance'),
                'transaction_category': order.get('PostingCategory'),
                'order_id': order.get('OrderId'),
                'market_id': order.get('MarketId'),
                'transaction_id': order.get('TransactionId')
            } for order in posts.get('Orders', {}).get('Order', []) if posts.get('Orders')]
    } if posts.get('Orders') and posts.get('Orders', {}).get('Order', []) else {}


def parse_account_balance(bal):
    return {
        'currency': bal.get('Currency'),
        'available_funds': bal.get('AvailableFunds'),
        'balance': bal.get('Balance'),
        'credit': bal.get('Credit'),
        'exposure': bal.get('Exposure'),
    }




import datetime
from requests import ConnectionError
from zeep.helpers import serialize_object

from betdaq.exceptions import APIError
from betdaq.utils import check_status_code, make_tz_naive


class BaseEndpoint(object):

    def __init__(self, parent):
        """
        :param parent: API client.
        """
        self.client = parent

    def request(self, method, params, secure=False):
        """
        :param method: The endpoint to be requested.
        :param params: Params to be used in request.
        :param secure: Whether the method belongs to the secure or readonly service.
        """
        try:
            if secure:
                response = self.client.secure_client.service[method](params)
            else:
                response = self.client.readonly_client.service[method](params)
        except ConnectionError:
            raise APIError(None, method, params, 'ConnectionError')
        except Exception as e:
            raise APIError(None, method, params, e)
        data = serialize_object(response)
        check_status_code(data)
        return data

    @staticmethod
    def process_response(response, date_time_sent, result_target, error_handler=None):
        """
        :param response: Response from request
        :param date_time_sent: Date time sent
        :param error_handler: function to parse _raw_elements from zeep response.
        :param result_target: name of the key to get response data from, changes per endpoint.
        """
        date_time_received = make_tz_naive(response.get('Timestamp')) or datetime.datetime.utcnow()
        if error_handler and response.get('_raw_elements'):
            response = error_handler(response)
        return {
            'data': response.get(result_target, []) if result_target else response,
            'date_time_sent': date_time_sent,
            'date_time_received': date_time_received,
        }




import os
from setuptools import setup, find_packages
from betdaq import __version__

setup(
    name="betdaq",
    version=__version__,
    author="Rory Cole",
    author_email="rory.cole1990@gmail.com",
    description="Betdaq API Python wrapper",
    url="https://github.com/rozzac90/betdaq_py",
    packages=find_packages(),
    install_requires=[line.strip() for line in open("requirements.txt")],
    long_description=open(os.path.join(os.path.dirname(__file__), 'README.md')).read(),
)
